import json
import os
from typing import Dict, List

import boto3
from botocore.exceptions import ClientError
from dotenv import load_dotenv

from tommytomato_utils.load_secrets.environment import Environment
from tommytomato_utils.load_secrets.exceptions import (
    ErrorWhenReadingInSecretsFromAWSSecretsManagerError
)


class SecretsLoader:

    def __init__(self, environment: Environment):
        self.environment = environment
        self.load_env_files()

    def load_env_files(self):
        # Load the base .env file
        base_dir = os.path.abspath(os.path.dirname(__name__))
        env_path = os.path.join(base_dir, 'credentials/.env')

        load_dotenv(env_path)

    def validate_secrets(self, secrets: Dict[str, str], required_secrets: List[str]) -> None:
        missing_secrets = [
            key for key in required_secrets if key not in secrets or secrets[key] is None
        ]
        if missing_secrets:
            raise ValueError(f"The following secrets are missing: {', '.join(missing_secrets)}")

    def load_from_env(self, required_secrets: List[str]) -> Dict[str, str]:
        secrets = {key: os.getenv(key)
                   for key in required_secrets}
        return secrets

    def load_from_aws(self, required_secrets: List[str]) -> Dict[str, str]:
        secret_name = f"{self.environment.name.lower()}/ops-tooling"
        region_name = "eu-central-1"

        # Create a Secrets Manager client
        session = boto3.session.Session()
        client = session.client(service_name='secretsmanager', region_name=region_name)

        try:
            get_secret_value_response = client.get_secret_value(SecretId=secret_name)
        except ClientError as exception:
            raise ErrorWhenReadingInSecretsFromAWSSecretsManagerError(
                secret_name,
                region_name,
                exception,
            )

        secret = get_secret_value_response['SecretString']
        aws_secrets = json.loads(secret)

        # Only return the required secrets
        secrets = {key: aws_secrets.get(key)
                   for key in required_secrets}
        return secrets

    def load_secrets(self, required_secrets: List[str]) -> Dict[str, str]:
        # Load secrets from .env
        secrets = self.load_from_env(required_secrets)

        # Check for missing secrets
        missing_secrets = [key for key, value in secrets.items() if value is None]

        if missing_secrets:
            # Load missing secrets from AWS
            aws_secrets = self.load_from_aws(missing_secrets)
            secrets.update(aws_secrets)

        # Validate all required secrets are now present
        self.validate_secrets(secrets, required_secrets)
        return secrets



import atexit
import logging
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Type

import pandas as pd
import sqlalchemy
from sqlalchemy import MetaData, Table, create_engine, exc, text
from sqlalchemy.orm import DeclarativeMeta, sessionmaker
from tenacity import retry, stop_after_attempt, wait_fixed

from tommytomato_utils.database_client.exceptions import (
    ColumnMismatchError, DatabaseConnectionError, NoValidPasswordProvidedError,
    SchemaAlreadyExistsError, TableNotFoundInReflectedDatabaseSchemaError
)

# Configure the logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


@dataclass
class DatabaseClientConfig:
    host: str
    port: int
    user: str
    password: str
    database: str

    def __post_init__(self):
        if not self.password or self.password == 'None':
            raise NoValidPasswordProvidedError(self.password)

    def get_connection_string(self):
        return f"postgresql://{self.user}:{self.password}@{self.host}:{self.port}/{self.database}"


class DatabaseClient:

    def __init__(self, config: DatabaseClientConfig, base: Optional[Type[DeclarativeMeta]] = None):
        self.engine = create_engine(
            config.get_connection_string(),
            pool_pre_ping=True,
            pool_size=1,
            max_overflow=1,
            pool_recycle=1800
        )
        self.Session = sessionmaker(bind=self.engine)
        self.metadata = MetaData()
        self.tables = {}
        self.base = base
        atexit.register(self.close_engine)

        try:
            self.test_connection()
        except exc.SQLAlchemyError as e:
            raise DatabaseConnectionError(str(e))

        self.reflect_schema()

    def test_connection(self):
        with self.engine.connect() as connection:
            connection.execute(text("SELECT 1"))

    def reflect_schema(self):
        self.metadata.reflect(bind=self.engine)
        for table_name, table in self.metadata.tables.items():
            self.tables[table_name] = table
        logger.info("\n\nSchema reflected:\n %s", self.tables.keys())

    def refresh_schema(self):
        self.metadata.clear()
        self.reflect_schema()

    @contextmanager
    def session_scope(self):
        session = self.Session()
        try:
            yield session
            session.commit()
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()

    def create_tables(self):
        if self.base is None:
            raise ValueError("The 'base' parameter must be provided when creating tables.")
        try:
            self._check_if_tables_exist(self.base)
            self.base.metadata.create_all(self.engine)
            logger.info("\nTables created successfully")
            self.refresh_schema()
        except exc.SQLAlchemyError as e:
            logger.error(f"An error occurred while creating tables: {e}")
            raise

    def _check_if_tables_exist(self, base: Type[DeclarativeMeta]):
        existing_tables = set(self.tables.keys())
        new_tables = set(base.metadata.tables.keys())
        intersecting_tables = existing_tables.intersection(new_tables)
        if intersecting_tables:
            raise SchemaAlreadyExistsError(list(intersecting_tables))

    def _validate_and_format_data_before_insert(self, table: Table, data: List[Dict[str, Any]]):
        df = pd.DataFrame(data)
        self._validate_columns(df, table)
        df = self._order_columns(df, table)
        data = df.to_dict(orient='records')
        return data

    def _validate_columns(self, df: pd.DataFrame, table: Table):
        expected_columns = self.tables[table.name].columns.keys()
        df_columns = df.columns.tolist()
        if not all(col in df_columns for col in expected_columns):
            missing_columns = [col for col in expected_columns if col not in df_columns]
            raise ColumnMismatchError(table=table, missing_columns=missing_columns)

    def _order_columns(self, df: pd.DataFrame, table: Table) -> pd.DataFrame:
        table_columns = self.tables[table.name].columns.keys()
        return df[table_columns]

    def insert_data(self, table_name: str, data: List[Dict[str, Any]]):
        if table_name not in self.tables:
            raise TableNotFoundInReflectedDatabaseSchemaError(table_name=table_name)
        table = self.tables[table_name]
        data = self._validate_and_format_data_before_insert(table, data)
        with self.session_scope() as session:
            session.execute(table.insert(), data)

    @retry(stop=stop_after_attempt(3), wait=wait_fixed(2))
    def query_data(self, query):
        if not isinstance(query, str):
            raise ValueError("The query parameter must be a string")
        with self.session_scope() as session:
            try:
                result = session.execute(text(query))
                df = pd.DataFrame(result.fetchall(), columns=list(result.keys()))
                return df
            except sqlalchemy.exc.ProgrammingError as exception:
                logger.error(f"An error occurred while querying data: {exception}")

    def execute_sql(self, sql, params=None):
        with self.session_scope() as session:
            try:
                if params:
                    result = session.execute(text(sql), params)
                else:
                    result = session.execute(text(sql))
                if result.returns_rows:
                    df = pd.DataFrame(result.fetchall(), columns=list(result.keys()))
                    return df
            except exc.SQLAlchemyError as e:
                logger.error(f"An error occurred while executing SQL: {e}")
                raise

    def close_engine(self):
        logger.debug("Closing engine...")
        self.engine.dispose()



import logging

from tommytomato_utils.logger.log_status import LogStatus


class Logger:

    def __init__(self, name: str = "tommytomato_operation_utils", level: int = logging.INFO):
        """
        Initializes the Logger.

        Args:
            name (str): Name of the logger.
            level (int): Logging level.
        """
        self.logger = logging.getLogger(name)
        self.logger.setLevel(level)
        self._setup_console_handler(level)

    def _setup_console_handler(self, level: int):
        """
        Sets up the console handler with the specified logging level.

        Args:
            level (int): Logging level.
        """
        console_handler = logging.StreamHandler()
        console_handler.setLevel(level)
        formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
        console_handler.setFormatter(formatter)
        self.logger.addHandler(console_handler)

    def add_handler(self, handler: logging.Handler):
        """
        Adds a logging handler to the logger.

        Args:
            handler (logging.Handler): Logging handler to add.
        """
        self.logger.addHandler(handler)

    def get_logger(self):
        """
        Returns the logger instance.
        """
        return self.logger

    def log(self, status: LogStatus, message: str):
        """
        Logs a message with the given status.

        Args:
            status (LogStatus): Logging status.
            message (str): Log message.
        """
        formatted_message = f"{status.value}: {message}"
        extra = {
            'log_status': status.value
        }
        self.logger.log(logging.INFO, formatted_message, extra=extra)



from typing import List

from pandas import DataFrame


class HashingError(Exception):
    pass


class InvalidInputTypeError(HashingError):

    def __init__(self, input_data):
        super().__init__(
            f"Invalid input type: {type(input_data).__name__}. "
            f"Expected list or DataFrame."
        )


class MissingHashColumnsError(HashingError):

    def __init__(self, dataframe: DataFrame, hash_columns: List[str]):
        missing_columns = [col for col in hash_columns if col not in dataframe.columns]
        super().__init__(
            '\nMissing columns for hash generation.'
            f'\nDataFrame columns: {dataframe.columns.tolist()}\n\n'
            f'Missing columns: {missing_columns}\n\n'
        )


class DuplicateHashValuesError(Exception):

    def __init__(self, duplicate_dataframe: DataFrame):
        super().__init__(
            f"\n Error: Duplicate hash values found, while allow_duplicates is set to False."
            f"Duplicate values:"
            f"\n\n{duplicate_dataframe}\n\n"
            f"Please ensure that all input datapoints are unique!\n"
        )



from enum import Enum
from typing import List


class Environment(Enum):
    """Contains all valid Environments"""
    TESTING = "TESTING"
    ACCEPTANCE = "ACCEPTANCE"
    PRODUCTION = "PRODUCTION"

    def __str__(self) -> str:
        return str(self.value)

    @classmethod
    def possible_environment_values(cls) -> List[str]:
        return [environment.value for environment in Environment]

    @classmethod
    def from_str(cls, env_str: str) -> 'Environment':
        if env_str is None:
            raise ValueError(
                "Environment string is None. "
                f"Valid environments are: {cls.possible_environment_values()}"
            )

        for env in cls:
            if env.value.lower() == env_str.lower():
                return env

        raise ValueError(
            f"Invalid environment: {env_str}. "
            f"Valid environments are: {cls.possible_environment_values()}"
        )



from datetime import date, datetime
from logging import Handler, error
from uuid import uuid4

from pytz import timezone

from tommytomato_utils.database_client.database_client import DatabaseClient


class DatabaseLoggingHandler(Handler):

    def __init__(
        self, db_client: DatabaseClient, hub_id: str, run_id: str, user_id: str, tool_name: str,
        production_date: date
    ):
        super().__init__()
        self.db_client = db_client
        self.hub_id = hub_id
        self.run_id = run_id
        self.user_id = user_id
        self.tool_name = tool_name
        self.production_date = production_date

    def emit(self, record):
        try:
            log_status = record.__dict__.get('log_status', 'status_undefined')
        except AttributeError:
            log_status = 'status_undefined'

        log_entry = {
            'id': str(uuid4()),
            'tool_name': self.tool_name,
            'status': log_status,
            'message': record.getMessage(),
            'run_id': self.run_id,
            'user_id': self.user_id,
            'hub_id': self.hub_id,
            'production_date': self.production_date,
            'inserted_at': datetime.now(timezone('Europe/Amsterdam')),
        }
        try:
            self.db_client.insert_data('ecs_task_logs', [log_entry])
        except Exception as e:
            error(f"Failed to log to database: {e}")



import datetime
import logging
import unittest
from unittest.mock import MagicMock

from tommytomato_utils.database_client.database_client import DatabaseClient
from tommytomato_utils.logger.configure_logging import configure_logging
from tommytomato_utils.logger.log_status import LogStatus


class TestLogger(unittest.TestCase):

    def setUp(self):
        self.db_client = MagicMock(DatabaseClient)
        self.hub_id = "hub123"
        self.run_id = "run456"
        self.user_id = "user789"
        self.tool_name = "my_tool"

    def test_logger_without_db_logging(self):
        custom_logger = configure_logging(log_level=logging.DEBUG)

        with self.assertLogs("tommytomato_operation_utils", level='DEBUG') as cm:
            custom_logger.log(LogStatus.STARTED, "Task has started without DB logging.")
            custom_logger.log(LogStatus.COMPLETED, "Task has completed without DB logging.")

        self.assertIn(
            "INFO:tommytomato_operation_utils:started: Task has started without DB logging.",
            cm.output
        )
        self.assertIn(
            "INFO:tommytomato_operation_utils:completed: Task has completed without DB logging.",
            cm.output
        )

    def test_logger_with_db_logging(self):
        custom_logger = configure_logging(
            db_client=self.db_client,
            hub_id=self.hub_id,
            run_id=self.run_id,
            user_id=self.user_id,
            tool_name=self.tool_name,
            production_date=datetime.date(year=2024, month=10, day=10),
            log_to_db=True,
            log_level=logging.DEBUG
        )

        with self.assertLogs("tommytomato_operation_utils", level='DEBUG') as cm:
            custom_logger.log(LogStatus.STARTED, "Task has started with DB logging.")
            custom_logger.log(LogStatus.COMPLETED, "Task has completed with DB logging.")

        self.assertIn(
            "INFO:tommytomato_operation_utils:started: Task has started with DB logging.",
            cm.output
        )
        self.assertIn(
            "INFO:tommytomato_operation_utils:completed: Task has completed with DB logging.",
            cm.output
        )

        custom_logger.log(LogStatus.STARTED, "Task has started with DB logging.")
        custom_logger.log(LogStatus.COMPLETED, "Task has completed with DB logging.")
        self.assertEqual(len(self.db_client.insert_data.call_args_list), 2)

        # Verify the first log entry
        args, kwargs = self.db_client.insert_data.call_args_list[0]
        log_entry = args[1][0]
        self.assertEqual(log_entry['hub_id'], self.hub_id)
        self.assertEqual(log_entry['run_id'], self.run_id)
        self.assertEqual(log_entry['user_id'], self.user_id)
        self.assertEqual(log_entry['tool_name'], self.tool_name)
        self.assertEqual(log_entry['status'], LogStatus.STARTED.value)
        self.assertEqual(log_entry['message'], "started: Task has started with DB logging.")

        # Verify the second log entry
        args, kwargs = self.db_client.insert_data.call_args_list[1]
        log_entry = args[1][0]
        self.assertEqual(log_entry['status'], LogStatus.COMPLETED.value)
        self.assertEqual(log_entry['message'], "completed: Task has completed with DB logging.")

    def test_missing_db_params(self):
        with self.assertRaises(ValueError):
            configure_logging(db_client=self.db_client, log_to_db=True, log_level=logging.DEBUG)



import datetime
import logging

from tommytomato_utils.database_client.database_client import DatabaseClient
from tommytomato_utils.logger.db_logging_handler import DatabaseLoggingHandler
from tommytomato_utils.logger.logger import Logger


def configure_logging(
    db_client: DatabaseClient = None,
    hub_id: str = None,
    run_id: str = None,
    user_id: str = None,
    tool_name: str = None,
    production_date: datetime.date = None,
    log_to_db: bool = False,
    log_level: int = logging.INFO
):
    """
    Configures the logging setup.

    Args:
        db_client (DatabaseClient): Database client for logging to the database.
        hub_id (str): Hub ID for logging.
        run_id (str): Run ID for logging.
        user_id (str): User ID for logging.
        tool_name (str): Tool name for logging.
        log_to_db (bool): Flag to enable logging to the database.
        log_level (int): Logging level.
        production_date (datetime.date) Production Date for logging.
    """
    custom_logger = Logger(level=log_level)

    if log_to_db:
        if not all([db_client, hub_id, run_id, user_id, tool_name]):
            raise ValueError(
                "All database logging parameters must be provided when log_to_db is True"
            )

        db_handler = DatabaseLoggingHandler(
            db_client=db_client,
            hub_id=hub_id,
            run_id=run_id,
            user_id=user_id,
            tool_name=tool_name,
            production_date=production_date
        )
        db_handler.setLevel(log_level)
        custom_logger.add_handler(db_handler)

    return custom_logger



from enum import Enum


class LogStatus(Enum):
    """
    Enum class for the status of a log message.
    """
    STARTED = "started"
    COMPLETED = "completed"
    FAILED = "failed"
    IN_PROGRESS = "in_progress"



import unittest
from unittest.mock import patch

from tommytomato_utils.load_secrets.environment import Environment
from tommytomato_utils.load_secrets.secrets_loader import SecretsLoader


class TestSecretsLoader(unittest.TestCase):

    def setUp(self):
        self.environment = Environment.TESTING

    @patch('tommytomato_utils.load_secrets.secrets_loader.load_dotenv')
    def test_load_env_files(self, mock_load_dotenv):
        secrets_loader = SecretsLoader(self.environment)
        secrets_loader.load_env_files()
        mock_load_dotenv.assert_called()

    @patch('os.getenv')
    def test_load_from_env(self, mock_getenv):
        secrets_loader = SecretsLoader(self.environment)
        required_secrets = ['SECRET_KEY']
        mock_getenv.return_value = 'mocked_secret_value'
        secrets = secrets_loader.load_from_env(required_secrets)
        self.assertEqual(secrets['SECRET_KEY'], 'mocked_secret_value')

    @patch('boto3.session.Session.client')
    def test_load_from_aws(self, mock_boto_client):
        secrets_loader = SecretsLoader(self.environment)
        required_secrets = ['AWS_SECRET_KEY']
        mock_client_instance = mock_boto_client.return_value
        mock_client_instance.get_secret_value.return_value = {
            'SecretString': '{"AWS_SECRET_KEY": "aws_mocked_secret_value"}'
        }
        secrets = secrets_loader.load_from_aws(required_secrets)
        self.assertEqual(secrets['AWS_SECRET_KEY'], 'aws_mocked_secret_value')

    def test_validate_secrets(self):
        secrets_loader = SecretsLoader(self.environment)
        secrets = {
            'SECRET_KEY': 'mocked_secret_value'
        }
        required_secrets = ['SECRET_KEY']
        secrets_loader.validate_secrets(secrets, required_secrets)

    def test_validate_secrets_missing(self):
        secrets_loader = SecretsLoader(self.environment)
        secrets = {}
        required_secrets = ['SECRET_KEY']
        with self.assertRaises(ValueError):
            secrets_loader.validate_secrets(secrets, required_secrets)



import unittest
from unittest.mock import MagicMock, patch

import pandas as pd
from sqlalchemy import Column, Integer, MetaData, String, Table, create_engine
from sqlalchemy.orm import declarative_base, sessionmaker

from tommytomato_utils.database_client.database_client import DatabaseClient, DatabaseClientConfig
from tommytomato_utils.database_client.exceptions import ColumnMismatchError


class TestDatabaseClient(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.config = DatabaseClientConfig(
            host='localhost',
            port=5432,
            user='test_user',
            password='test_password',
            database='test_db'
        )
        cls.Base = declarative_base()
        cls.Base.metadata.clear()  # Clear metadata to prevent conflicts

        # Define a test table within the Base
        class TestTable(cls.Base):
            __tablename__ = 'test_table'
            id = Column(Integer, primary_key=True)
            name = Column(String)
            age = Column(Integer)

    @patch('tommytomato_utils.database_client.database_client.create_engine')
    def setUp(self, mock_create_engine):
        # Mocking the create_engine method to use an in-memory SQLite database for testing
        self.engine = create_engine('sqlite:///:memory:')
        mock_create_engine.return_value = self.engine
        self.db_client = DatabaseClient(self.config, base=self.Base)
        self.metadata = MetaData()
        # Create a test table
        self.test_table = Table(
            'test_table', self.metadata, Column('id', Integer, primary_key=True),
            Column('name', String), Column('age', Integer)
        )
        self.metadata.create_all(self.engine)
        # Manually add the test table to the db_client's tables dictionary
        self.db_client.tables['test_table'] = self.test_table
        # Creating a sessionmaker for the test class
        self.Session = sessionmaker(bind=self.engine)

    def tearDown(self):
        self.metadata.drop_all(self.engine)

    def test_connection(self):
        try:
            self.db_client.test_connection()
        except Exception as e:
            self.fail(f"test_connection raised Exception unexpectedly: {e}")

    def test_reflect_schema(self):
        self.db_client.reflect_schema()
        self.assertIn('test_table', self.db_client.tables)
        self.assertEqual(self.db_client.tables['test_table'].columns.keys(), ['id', 'name', 'age'])

    def test_validate_columns(self):
        df = pd.DataFrame({
            'id': [1],
            'name': ['John'],
            'age': [30]
        })
        try:
            self.db_client._validate_columns(df, self.test_table)
        except Exception as e:
            self.fail(f"_validate_columns raised Exception unexpectedly: {e}")

        df_invalid = pd.DataFrame({
            'id': [1],
            'name': ['John']
        })
        with self.assertRaises(ColumnMismatchError):
            self.db_client._validate_columns(df_invalid, self.test_table)

    def test_order_columns(self):
        df = pd.DataFrame({
            'name': ['John'],
            'age': [30],
            'id': [1]
        })
        ordered_df = self.db_client._order_columns(df, self.test_table)
        self.assertEqual(list(ordered_df.columns), ['id', 'name', 'age'])

    def test_insert_data(self):
        data = [{
            'id': 1,
            'name': 'John',
            'age': 30
        }]
        try:
            self.db_client.insert_data('test_table', data)
        except Exception as e:
            self.fail(f"insert_data raised Exception unexpectedly: {e}")

        with self.engine.connect() as connection:
            result = connection.execute(self.test_table.select()).fetchall()
            self.assertEqual(len(result), 1)
            self.assertEqual(result[0], (1, 'John', 30))

    def test_query_data(self):
        data = [{
            'id': 1,
            'name': 'John',
            'age': 30
        }]
        self.db_client.insert_data('test_table', data)

        query = 'SELECT * FROM test_table'
        df = self.db_client.query_data(query)
        self.assertEqual(df.shape[0], 1)
        self.assertEqual(df.iloc[0]['name'], 'John')

    def test_execute_sql(self):
        data = [{
            'id': 1,
            'name': 'John',
            'age': 30
        }]
        self.db_client.insert_data('test_table', data)

        sql = 'SELECT * FROM test_table WHERE id = :id'
        params = {
            'id': 1
        }
        df = self.db_client.execute_sql(sql, params)
        self.assertEqual(df.shape[0], 1)
        self.assertEqual(df.iloc[0]['name'], 'John')

    def test_query_with_params(self):
        # Insert test data
        data = [{
            'id': 1,
            'name': 'John',
            'age': 30
        }, {
            'id': 2,
            'name': 'Jane',
            'age': 25
        }]
        self.db_client.insert_data('test_table', data)

        # Define the query with parameters
        query = 'SELECT * FROM test_table WHERE age = :age'
        params = {
            'age': 30
        }

        # Execute the query with parameters
        df = self.db_client.execute_sql(query, params)
        self.assertEqual(df.shape[0], 1)
        self.assertEqual(df.iloc[0]['name'], 'John')
        self.assertEqual(df.iloc[0]['age'], 30)

    @patch('tommytomato_utils.database_client.database_client.create_engine')
    @patch(
        'tommytomato_utils.database_client.database_client.DatabaseClient.reflect_schema',
        MagicMock()
    )
    @patch(
        'tommytomato_utils.database_client.database_client.DatabaseClient.test_connection',
        MagicMock()
    )
    def test_create_tables_without_base(self, mock_create_engine):
        mock_create_engine.return_value = create_engine('sqlite:///:memory:')
        db_client_no_base = DatabaseClient(self.config)
        with self.assertRaises(ValueError):
            db_client_no_base.create_tables()

    def test_create_tables_with_base(self):
        # Drop the test table to simulate table creation
        self.metadata.drop_all(self.engine)
        self.db_client.tables.pop('test_table', None)

        try:
            self.db_client.create_tables()
        except Exception as e:
            self.fail(f"create_tables raised Exception unexpectedly: {e}")



from unittest import TestCase

from pandas import DataFrame
from pandas.testing import assert_frame_equal
from typeguard import TypeCheckError

from tommytomato_utils.hashing_client.exceptions import (
    DuplicateHashValuesError, MissingHashColumnsError
)
from tommytomato_utils.hashing_client.hashing_client import HashingClient


class TestHashGenerator(TestCase):

    def test_hashing_string(self):
        input_string = 'hello'
        expected_output = '9342d47a-1bab-5709-9869-c840b2eac501'
        output = HashingClient.get_hash_uuid_from_string(input_string)
        self.assertEqual(output, expected_output)

    def test_dataframe_hash_with_just_columns_needed(self):
        df = DataFrame({
            'col1': ['value1', 'value2'],
            'col2': ['value3', 'value4'],
        })
        hash_columns = ['col1', 'col2']

        expected_output = df.copy()
        expected_output['box_id'] = [
            '1f3ac418-9828-51c9-9404-f88123f7fa65', 'f9171cd3-e34d-50c2-993d-803a16ace55f'
        ]

        output = HashingClient.add_dataframe_column_hash_given_column_names(
            df, hash_columns, 'box_id'
        )
        assert_frame_equal(output, expected_output)

    def test_dataframe_hash_with_additional_column(self):
        """Same as previous test, but adding a new column that has no impact on output."""
        df = DataFrame(
            {
                'col1': ['value1', 'value2'],
                'col2': ['value3', 'value4'],
                'col3': ['value5', 'value6'],
            }
        )
        hash_columns = ['col1', 'col2']

        expected_output = df.copy()
        expected_output['box_id'] = [
            '1f3ac418-9828-51c9-9404-f88123f7fa65', 'f9171cd3-e34d-50c2-993d-803a16ace55f'
        ]

        output = HashingClient.add_dataframe_column_hash_given_column_names(
            df, hash_columns, 'box_id'
        )
        assert_frame_equal(output, expected_output)

    def test_missing_columns_error(self):
        df = DataFrame({
            'col1': ['value1', 'value2'],
        })
        hash_columns = ['col1', 'col2']  # Col 2 not in dataframe!

        with self.assertRaises(MissingHashColumnsError):
            HashingClient.add_dataframe_column_hash_given_column_names(df, hash_columns, 'box_id')

    def test_wrong_input_type(self):
        df = None
        with self.assertRaises(TypeCheckError):
            HashingClient.add_dataframe_column_hash_given_column_names(df, ['col1'], 'box_id')

    def test_no_duplicates_allowed_raises_error(self):
        df = DataFrame({
            'col1': ['value1', 'value1'],
            'col2': ['value3', 'value3'],
        })
        hash_columns = ['col1', 'col2']

        with self.assertRaises(DuplicateHashValuesError):
            HashingClient.add_dataframe_column_hash_given_column_names(
                df, hash_columns, 'box_id', allow_duplicates=False
            )

    def test_no_duplicates_allowed_passes(self):
        df = DataFrame({
            'col1': ['value1', 'value2'],
            'col2': ['value3', 'value4'],
        })
        hash_columns = ['col1', 'col2']

        expected_output = df.copy()
        expected_output['box_id'] = [
            '1f3ac418-9828-51c9-9404-f88123f7fa65', 'f9171cd3-e34d-50c2-993d-803a16ace55f'
        ]

        output = HashingClient.add_dataframe_column_hash_given_column_names(
            df, hash_columns, 'box_id', allow_duplicates=False
        )
        assert_frame_equal(output, expected_output)



from typing import List
from uuid import NAMESPACE_DNS, uuid5

from pandas import DataFrame, Series
from typeguard import typechecked

from tommytomato_utils.hashing_client.constants import HASHING_FIELD_SEPARATOR
from tommytomato_utils.hashing_client.exceptions import \
    DuplicateHashValuesError  # New custom exception
from tommytomato_utils.hashing_client.exceptions import MissingHashColumnsError


class HashingClient:
    """
    Client to generate a hash based on either a list of string values or specific columns of a
    DataFrame.
    """

    @typechecked
    @staticmethod
    def get_hash_uuid_from_string(string_value: str) -> str:
        string_value = str(string_value)
        output_hash = str(uuid5(NAMESPACE_DNS, string_value))
        return output_hash

    @typechecked
    @staticmethod
    def add_dataframe_column_hash_given_column_names(
        dataframe: DataFrame,
        hash_column_names: List[str],
        column_name: str,
        allow_duplicates: bool = True,
    ) -> DataFrame:
        try:
            concatenated_columns = Series(
                map(
                    HASHING_FIELD_SEPARATOR.join,
                    dataframe[hash_column_names].astype(str).values.tolist(),
                ),
                index=dataframe.index,
            )
        except KeyError:
            raise MissingHashColumnsError(dataframe, hash_column_names)

        if not allow_duplicates:
            duplicate_values = concatenated_columns[concatenated_columns.duplicated()]
            if not duplicate_values.empty:
                raise DuplicateHashValuesError(duplicate_values)

        dataframe[column_name] = concatenated_columns.apply(
            HashingClient.get_hash_uuid_from_string
        )
        return dataframe



class ErrorWhenReadingInSecretsFromAWSSecretsManagerError(Exception):

    def __init__(self, secret_name: str, region_name: str, exception: Exception):
        super().__init__(
            "An error occurred when reading in secrets from AWS Secrets Manager. "
            f"\nThe following error occurred: {exception}\n\n"
            "Secrets are gathered from:\n\n"
            f"-secret_name: {secret_name}\n"
            f"-region_name: {region_name}\n"
        )



from setuptools import find_packages, setup

requires = [
    'boto3==1.34.130',
    'pandas==2.2.2',
    'sqlalchemy==2.0.31',
    'tenacity==8.5.0',
    'python-dotenv==0.18.0',
    'psycopg2-binary==2.9.9',
    'typeguard==4.3.0',  # to enforce types
]

linting_requires = ['flake8==7.1.0', 'isort==5.10.1', 'pre-commit', 'importlib_metadata==4.8.3']

testing_requires = [
    'pytest',
    'pytest-cov',
    'freezegun',
]

setup(
    name='tommytomato_operation_utils',
    version='1.4.2',
    description='Utility package for hash generation and database operations.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='felix_bithero',
    author_email='felixl@bithero.com',
    url='https://github.com/felix_bithero/tommytomato_operation_utils',
    packages=find_packages(),
    install_requires=requires,
    extras_require={
        'linting': linting_requires,
        'testing': testing_requires,
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.11',
)



from typing import List

from sqlalchemy import Table


class NoValidPasswordProvidedError(Exception):

    def __init__(self, password: str):
        super().__init__(
            "Improper configuration! "
            "You have not passed a valid password.\n"
            f"Password: '{password}'"
        )


class DatabaseConnectionError(Exception):

    def __init__(self, message: str):
        super().__init__(f"Failed to connect to the database: {message}")


class ColumnMismatchError(Exception):

    def __init__(self, table: Table, missing_columns: List[str]):
        super().__init__(
            f"Missing columns in DataFrame to upload to database for table '{table.name}':"
            f"{missing_columns}"
        )


class TableNotFoundInReflectedDatabaseSchemaError(Exception):

    def __init__(self, table_name: str):
        super().__init__(f"The table '{table_name}' is not available in the reflected schema")


class SchemaAlreadyExistsError(Exception):

    def __init__(self, table_names: List[str]):
        super(
        ).__init__(f"The schema already contains the following tables: {', '.join(table_names)}")
